import os
import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import datasets, transforms, models
from torch.utils.data import DataLoader
from torch.ao.quantization import quantize_dynamic

def main():
    # 設定
    data_dir = "train"
    batch_size = 32
    num_epochs = 5
    learning_rate = 0.001
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # データ前処理
    transform = transforms.Compose([
        transforms.Resize((224, 224)),
        transforms.ToTensor(),
        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ])

    # データセットとデータローダ
    train_dataset = datasets.ImageFolder(root=data_dir, transform=transform)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    num_classes = len(train_dataset.classes)

    # モデルのロードと構成変更
#    model = models.mobilenet_v2(pretrained=True)  # 非推奨
    model = models.mobilenet_v2(weights = models.MobileNet_V2_Weights.IMAGENET1K_V1)
    for param in model.parameters():
        param.requires_grad = False
    model.classifier[1] = nn.Linear(model.last_channel, num_classes)
    model = model.to(device)

    # 損失関数と最適化手法
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.classifier.parameters(), lr=learning_rate)

    # 学習ループ
    model.train()
    for epoch in range(num_epochs):
        running_loss = 0.0
        for inputs, labels in train_loader:
            inputs, labels = inputs.to(device), labels.to(device)

            optimizer.zero_grad()
            outputs = model(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            running_loss += loss.item()

        print(f"Epoch [{epoch+1}/{num_epochs}] Loss: {running_loss/len(train_loader):.4f}")

    # 量子化（動的）
    model.cpu()
    quantized_model = quantize_dynamic(
        model,
        {nn.Linear},
        dtype=torch.qint8
    )

    # モデルの保存
    save_path = "mobilenetv2_quantized.pth"
    torch.save(quantized_model.state_dict(), save_path)
    print(f"量子化済みモデルを '{save_path}' に保存しました。")

if __name__ == "__main__":
    main()
